/*
 *  Copyright (c) 2008 Cyrille Berger <cberger@cberger.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * either version 2, or (at your option) any later version of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "Library.h"
#include "Library_p.h"

#include <sstream>
#include <fstream>

#include <llvm/Module.h>
#include <llvm/ModuleProvider.h>

#include "GTLCore/ErrorMessage.h"
#include "GTLCore/Function.h"
#include "GTLCore/ModuleData_p.h"
#include "GTLCore/PixelDescription.h"
#include "GTLCore/Region.h"
#include "GTLCore/Type.h"
#include "GTLCore/TypeManager.h"
#include "GTLCore/Value.h"
#include "GTLCore/VirtualMachine_p.h"
#include "GTLCore/wrappers/Allocate.h"

#include "Debug.h"
#include "CodeGenerator_p.h"
#include "Compiler_p.h"
#include "Wrapper_p.h"
#include "wrappers/ImageWrap_p.h"
#include "wrappers/RegionWrap_p.h"

#include "Kernel_p.h"

using namespace OpenShiva;

Library::Library( bool _isKernel , int _channelsNb) : d(new Private)
{
  d->name = "";
  d->compiled = false;
  d->moduleProvider = 0;
  d->m_moduleData = 0;
  d->count_channels_generic = _channelsNb;
  d->isKernel = _isKernel;
  d->isStandardLibrary = false;
}

Library::~Library()
{
  cleanup();
  delete d;
}

void Library::cleanup()
{
  if(d->moduleProvider)
  {
    GTLCore::VirtualMachine::instance()->unregisterModule( d->moduleProvider);
    delete d->moduleProvider;
    d->moduleProvider = 0;
  }
  delete d->m_moduleData;
  d->m_moduleData = 0;
}


const GTLCore::String& Library::name() const
{
  return d->name;
}

void Library::setSource(const GTLCore::String& _source )
{
  d->source = _source;
}

void Library::loadFromFile(const GTLCore::String& _fileName)
{
  d->isStandardLibrary = _fileName.endWith( "shivastdlib.shiva" );
  d->source = "";
  std::ifstream in;
  in.open(_fileName.c_str() );
  if(not in)
  {
    SHIVA_DEBUG( "Impossible to open file " << _fileName );
    return;
  }
  GTLCore::String str;
  std::getline(in,str);
  while ( in ) {
    d->source += str;
    d->source += "\n";
    std::getline(in,str);
  }
}

void Library::postCompilation()
{
}

void Library::compile()
{
  if(d->source.empty()) return;
  cleanup();
  d->m_moduleData = new GTLCore::ModuleData(new llvm::Module(d->name));
  Compiler c( d->isKernel,  d->count_channels_generic );
  Wrapper::fillTypeManager( d->m_moduleData, d->m_moduleData->typeManager(), c.convertCenter() , d->count_channels_generic );
  GTLCore::String nameSpace;
  bool result = c.compile( not d->isStandardLibrary, d->source, d->name, d->m_moduleData, nameSpace );

  if(result)
  {
    d->compiled = true;
    // Register the compiled module in the virtual machine
    d->m_moduleData->doLink();
    d->moduleProvider = new llvm::ExistingModuleProvider( d->m_moduleData->llvmModule() );
    GTLCore::VirtualMachine::instance()->registerModule( d->moduleProvider );
    postCompilation();
    d->name = nameSpace;

  } else {
    cleanup();
    d->compilationErrors = c.errorMessages();
  }
  
}

GTLCore::String Library::compilationErrorsMessage() const
{
  std::ostringstream os;
  for( std::list<GTLCore::ErrorMessage>::iterator it = d->compilationErrors.begin();
       it != d->compilationErrors.end(); ++it)
  {
    os << it->fileName() << " at " << it->line() << " : " << it->errorMessage()  << std::endl;
  }
  return os.str();
}


bool Library::isCompiled() const
{
  return d->compiled;
}

const std::list<GTLCore::ErrorMessage>& Library::compilationErrors() const
{
  return d->compilationErrors;
}

GTLCore::String Library::asmSourceCode() const
{
  std::ostringstream os;
  os << *d->m_moduleData->llvmModule() << std::endl;
  return os.str();
}

const GTLCore::ModuleData* Library::data() const
{
  return d->m_moduleData;
}

std::list<GTLCore::Function*> Library::functions()
{
  return d->m_moduleData->functions();
}
